using Assets.Editor;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using UnityEditor;
using UnityEngine;

public class EditorExample : EditorWindow
{
    /// <summary>
    /// An Icon shown on the Button.
    /// </summary>
    private static Texture buttonIcon = null;

    /// <summary>
    /// Contains all Research Nodes shown in the Editor.
    /// </summary>
    private static List<ResearchNode> researchNodes = new List<ResearchNode>();

    /// <summary>
    /// The Windows Position and Size. Has to be saved between OnGUI()-Calls in order to make the Window draggable.
    /// </summary>
    private static List<Rect> windowPositionAndSizes = new List<Rect>();

    /// <summary>
    /// When setting a new Relation between two Research Nodes, the Source Node is saved here.
    /// </summary>
    private static ResearchNode linkResearchNodeSource = null;

    /// <summary>
    /// The Scroll-Position for the Scrollbar withing the ResearchNode Area.
    /// </summary>
    private Vector2 scrollPosition = Vector2.zero;

    /// <summary>
    /// Adds a Menu Item to the Window Menu.
    /// </summary>
    [MenuItem("Window/My Project/Research Editor Window")]
    static void InitEditorWindow()
    {
        //Show existing Window Instance. If one doesn't exist, make one
        EditorExample editor = EditorWindow.GetWindow<EditorExample>();

        //Give the Editor a nice Caption
        GUIContent titleContent = new GUIContent("Research Editor");
        editor.titleContent = titleContent;

        //Load Icons
        buttonIcon = ((Sprite)Resources.Load("Textures/ButtonIcon", typeof(Sprite))).texture;

        //Init Test Data
        GenerateExampleResearchNodes();
        GenerateWindowsForResearchNodes();
    }

    /// <summary>
    /// This Method is called whenever the UI is updated.
    /// </summary>
    void OnGUI()
    {
        EditorGUILayout.BeginHorizontal();
        {
            //Define Load-Button
            if (GUILayout.Button(new GUIContent("Load"), GUILayout.Width(150.0f)))
            {
                //Get the saved File
                string file = EditorUtility.OpenFilePanel("Load Techtree", "", "json");

                //Load Techtree from JSON
                if (!string.IsNullOrEmpty(file))
                {
                    LoadResearchNodes(file);
                    LoadWindowPositions(file);
                }
            }

            //Define Save-Button
            if (GUILayout.Button(new GUIContent("Save"), GUILayout.Width(150.0f)))
            {
                //Set the File to save
                string file = EditorUtility.SaveFilePanel("Save Techtree", "", "Techtree", "json");

                //Save Techtree as JSON
                if (!string.IsNullOrEmpty(file))
                {
                    SaveResearchNodes(file);
                    SaveWindowPositions(file);
                }
            }

            //Define NewResearchNode-Button
            if (GUILayout.Button(new GUIContent("New Research Node"), GUILayout.Width(150.0f)))
            {
                AddNewResearchNode();
            }
        }
        EditorGUILayout.EndHorizontal();

        //Start Zooming Area
        //EditorZoomArea.Begin(1, new Rect(0, 0, 5000, 5000));

        //Scrollbars
        scrollPosition = GUI.BeginScrollView(new Rect(0, 35 / 1, position.width / 1, (position.height - 35) / 1),
                                             scrollPosition,
                                             new Rect(0, 0, 5000, 5000));

        //Background Graph
        Rect graphPosition = new Rect(0, 0, 5000, 5000);
        GraphBackground.DrawGraphBackground(graphPosition, graphPosition);

        //Draw Lines between Nodes
        for (int i = 0; i < researchNodes.Count; i++)
        {
            //Draw a Line for each Relation
            for (int j = 0; j < researchNodes[i].researchNodeUnlockedResearchNodes.Count; j++)
            {
                //Offset the Endpoints Height, since Unity sets the Rect to a Height not considering the Windows' Heading
                Rect startWindow = windowPositionAndSizes[i];
                Rect targetWindow = windowPositionAndSizes[researchNodes.IndexOf(researchNodes[i].researchNodeUnlockedResearchNodes[j])];
                DrawLineBetweenWindows(startWindow, targetWindow);
            }
        }

        //Define all your Windows between BeginWindows() and EndWindows(). The { } Brackets are not needed, but make the Code easiert to read
        BeginWindows();
        {
            //Show a Window for all Research Nodes
            for (int i = 0; i < windowPositionAndSizes.Count; i++)
            {
                //Align Nodes to Grid
                Rect rect = windowPositionAndSizes[i];
                rect.x = ((int)(rect.x / 25.0f)) * 25.0f;
                rect.y = ((int)(rect.y / 25.0f)) * 25.0f;

                //Draw the Research Nodes
                windowPositionAndSizes[i] = GUILayout.Window(researchNodes[i].researchNodeID, rect, DrawResearchNodeWindow, researchNodes[i].researchNodeName);
            }
        }
        EndWindows();

        GUI.EndScrollView();

        //EditorZoomArea.End();
    }

    /// <summary>
    /// This Method defines UI-Elements within the Window.
    /// </summary>
    /// <param name="id">The Windows unique ID, which equals the ResearchNodes ID.</param>
    private void DrawResearchNodeWindow(int id)
    {
        //Get Array Index by ResearchNode ID
        int index = researchNodes.FindIndex(x => x.researchNodeID == id);

        //Define all UI-Elements for a single Research Node, stacked vertically
        GUILayout.BeginVertical();
        {
            //Button to delete the Research Node
            if (GUILayout.Button(new GUIContent("X"), GUILayout.Width(20.0f), GUILayout.Height(20.0f)))
            {
                DeleteResearchNode(id);
                return; //Return here to prevent remaining Code to cause Exceptions
            }

            //Nodes Name Edit-Fields
            GUILayout.Label("Name:", GUILayout.Width(220));
            researchNodes[index].researchNodeName = EditorGUILayout.TextField(researchNodes[index].researchNodeName, GUILayout.Width(220));

            GUILayout.Space(10);

            //Nodes ResearchPoints Edit-Fields
            GUILayout.Label("Research Points:", GUILayout.Width(220));
            researchNodes[index].researchNodeCost = EditorGUILayout.IntField(researchNodes[index].researchNodeCost, GUILayout.Width(220));

            GUILayout.Space(10);

            //Nodes related ResearchNodes
            GUILayout.Label("Related Research Nodes:", GUILayout.Width(220));
            for (int i = 0; i < researchNodes[index].researchNodeUnlockedResearchNodes.Count; i++)
            {
                GUILayout.BeginHorizontal();
                {
                    //ResearchNode Name Label
                    GUILayout.Label(researchNodes[index].researchNodeUnlockedResearchNodes[i].researchNodeName, GUILayout.Width(200));

                    //Button to delete Relation
                    if (GUILayout.Button("X", GUILayout.Width(20.0f), GUILayout.Height(20.0f)))
                    {
                        UnlinkTwoResearchNodes(researchNodes[index], researchNodes[index].researchNodeUnlockedResearchNodes[i]);
                        return; //Return here to prevent remaining Code to cause Exceptions
                    }
                }
                GUILayout.EndHorizontal();
            }

            //Button to add new Relation
            if (linkResearchNodeSource == null)
            {
                if (GUILayout.Button("Add new related Research Node", GUILayout.Width(220.0f), GUILayout.Height(20.0f)))
                {
                    //Start Linking Process by saving the Source Node
                    linkResearchNodeSource = researchNodes[index];
                    return; //Return here to prevent remaining Code to cause Exceptions
                }
            }

            //Button to finish the new Relation Process. Only show, if the Linking Process has been started and this is not the Source Node
            if (linkResearchNodeSource != null && linkResearchNodeSource != researchNodes[index])
            {
                if (GUILayout.Button("Link to this Node", GUILayout.Width(220.0f), GUILayout.Height(20.0f)))
                {
                    //Add Link
                    LinkTwoResearchNodes(linkResearchNodeSource, researchNodes[index]);

                    //Reset Process
                    linkResearchNodeSource = null;
                    return; //Return here to prevent remaining Code to cause Exceptions
                }
            }
        }
        GUILayout.EndVertical();

        //Set the Window as dragable
        GUI.DragWindow();
    }

    /// <summary>
    /// Generates some example Research Nodes, which we will use to test the Editor Tools.
    /// </summary>
    private static void GenerateExampleResearchNodes()
    {
        //Reset existing Data
        researchNodes.Clear();

        //Define some Research Nodes with unique IDs, Display Names and Costs
        researchNodes.Add(new ResearchNode() { researchNodeID = 0, researchNodeName = "Fire", researchNodeCost = 10 });
        researchNodes.Add(new ResearchNode() { researchNodeID = 1, researchNodeName = "Masonry", researchNodeCost = 25 });
        researchNodes.Add(new ResearchNode() { researchNodeID = 2, researchNodeName = "Writing", researchNodeCost = 25 });
        researchNodes.Add(new ResearchNode() { researchNodeID = 3, researchNodeName = "Architecture", researchNodeCost = 50 });
        researchNodes.Add(new ResearchNode() { researchNodeID = 4, researchNodeName = "Culture", researchNodeCost = 75 });

        //Link the Research Nodes together
        researchNodes[0].researchNodeUnlockedResearchNodes = new List<ResearchNode> { researchNodes[1], researchNodes[2] };
        researchNodes[1].researchNodeUnlockedResearchNodes = new List<ResearchNode> { researchNodes[3], researchNodes[4] };
    }

    /// <summary>
    /// Initializes the Window-Sizes and Positions for all Research Nodes.
    /// </summary>
    private static void GenerateWindowsForResearchNodes()
    {
        //Reset existing Data
        windowPositionAndSizes.Clear();

        //Set initial Window Locations
        for (int i = 0; i < researchNodes.Count; i++)
        {
            windowPositionAndSizes.Add(new Rect(new Vector2(100.0f * i + 50.0f, 50.0f * i + 50.0f), new Vector2(150.0f, 250.0f)));
        }
    }

    /// <summary>
    /// Calculates a Line between two Windows and draws it.
    /// The Line starts at the Top of the Start-Window and ends at the Bottom of the End-Window.
    /// </summary>
    /// <param name="start">The Start-Window Rectangle.</param>
    /// <param name="end">The End-Window Rectangle.</param>
    private void DrawLineBetweenWindows(Rect start, Rect end)
    {
        //Calculate Line Parameters
        Vector3 startPos = new Vector3(start.x + start.width / 2.0f, start.y, 0.0f);
        Vector3 endPos = new Vector3(end.x + end.width / 2.0f, end.y + end.height, 0.0f);
        Vector3 startTan = startPos + Vector3.down * 50.0f;
        Vector3 endTan = endPos + Vector3.up * 50.0f;
        Color shadowCol = new Color(0.0f, 1.0f, 1.0f, 0.06f);

        //Draw Line Shadows
        for (int i = 0; i < 3; i++)
        {
            Handles.DrawBezier(startPos, endPos, startTan, endTan, shadowCol, null, (i + 1) * 10.0f);
        }

        //Draw Lines
        Handles.DrawBezier(startPos, endPos, startTan, endTan, Color.cyan, null, 4.0f);
    }

    /// <summary>
    /// Loads a Set of ResearchNodes from a serialized JSON-Container.
    /// </summary>
    /// <param name="fileName">The FileName (including Path and Ending) to load the ResearchNodes from.</param>
    private void LoadResearchNodes(string fileName)
    {
        JsonParser<List<ResearchNode>> jsonParser = new JsonParser<List<ResearchNode>>();
        researchNodes = jsonParser.LoadData(fileName);
    }

    /// <summary>
    /// Saves the ResearchNodes as JSON using the provided FileName.
    /// </summary>
    /// <param name="fileName">The FileName (including Path and Ending) to save the ResearchNodes at.</param>
    private void SaveResearchNodes(string fileName)
    {
        JsonParser<List<ResearchNode>> jsonParser = new JsonParser<List<ResearchNode>>();
        jsonParser.SaveData(researchNodes, fileName);
    }

    /// <summary>
    /// Loads the Window Positions from a serialized JSON-Container.
    /// </summary>
    /// <param name="fileName">The FileName (including Path and Ending) to load the Window Positions from.</param>
    private void LoadWindowPositions(string fileName)
    {
        JsonParser<List<Tuple<float, float>>> jsonParser = new JsonParser<List<Tuple<float, float>>>();
        windowPositionAndSizes = jsonParser.LoadData($"{Path.GetFileNameWithoutExtension(fileName)}_pos.json").Select(x => new Rect(x.Item1, x.Item2, 150.0f, 250.0f)).ToList();
    }

    /// <summary>
    /// Saves the Window Positions as JSON using the provided FileName.
    /// </summary>
    /// <param name="fileName">The FileName (including Path and Ending) to save the Window Positions at.</param>
    private void SaveWindowPositions(string fileName)
    {
        List<Tuple<float, float>> windowPositions = windowPositionAndSizes.Select(x => new Tuple<float, float>(x.x, x.y)).ToList();

        JsonParser<List<Tuple<float, float>>> jsonParser = new JsonParser<List<Tuple<float, float>>>();
        jsonParser.SaveData(windowPositions, $"{Path.GetFileNameWithoutExtension(fileName)}_pos.json");
    }

    /// <summary>
    /// Adds a new empty Research Node.
    /// </summary>
    private void AddNewResearchNode()
    {
        researchNodes.Add(new ResearchNode() { researchNodeID = GetNextAvailableFreeResearchNodeID(), researchNodeName = "", researchNodeCost = 0 });

        windowPositionAndSizes.Add(new Rect(new Vector2(100.0f, 100.0f), new Vector2(150.0f, 250.0f)));
    }

    /// <summary>
    /// Returns the next unused Research Node ID.
    /// Trys to fill Gaps in the continuous Node IDs, if any Nodes have been deleted previously.
    /// </summary>
    /// <returns>A Research Node ID.</returns>
    private int GetNextAvailableFreeResearchNodeID()
    {
        //Look for a free Research Node ID
        for (int i = 0; i < int.MaxValue; i++)
        {
            if (researchNodes.Any(x => i == x.researchNodeID) == false)
            {
                return i;
            }
        }

        //Will never occur, but to be safe
        throw new Exception("Unable to find any free ID for new ResearchNode!");
    }

    /// <summary>
    /// Deletes the Research Node with the provided Research Node ID.
    /// </summary>
    /// <param name="id">The Research Nodes ID.</param>
    private void DeleteResearchNode(int id)
    {
        //Try to find the Research Node to delete
        int index = researchNodes.IndexOf(researchNodes.FirstOrDefault(x => id == x.researchNodeID));
        
        //Only proceed, if the Node was found
        if (index >= 0)
        {
            //Save the Node for now
            ResearchNode researchNode = researchNodes[index];

            //Remove Node and Window-Layout-Information
            researchNodes.RemoveAt(index);
            windowPositionAndSizes.RemoveAt(index);

            //Remove any Links from other Research Nodes
            for (int i = 0; i < researchNodes.Count; i++)
            {
                researchNodes[i].researchNodeUnlockedResearchNodes.Remove(researchNode);
            }

            //Done
            return;
        }

        //Will never occur, but to be safe
        throw new Exception($"Unable to delete ResearchNode with ID {id}!");
    }

    /// <summary>
    /// Links two Research Nodes.
    /// </summary>
    /// <param name="source">The Source Research Node.</param>
    /// <param name="destination">The Destination Research Node.</param>
    private void LinkTwoResearchNodes(ResearchNode source, ResearchNode destination)
    {
        //Don't add the Relation twice
        if (source.researchNodeUnlockedResearchNodes.Contains(destination) == false)
        {
            source.researchNodeUnlockedResearchNodes.Add(destination);
        }
    }

    /// <summary>
    /// Unlinks two Research Nodes.
    /// </summary>
    /// <param name="source">The Source Research Node.</param>
    /// <param name="destination">The Destination Research Node.</param>
    private void UnlinkTwoResearchNodes(ResearchNode source, ResearchNode destination)
    {
        //Only unlink, if available
        if (source.researchNodeUnlockedResearchNodes.Contains(destination))
        {
            source.researchNodeUnlockedResearchNodes.Remove(destination);
        }
    }
}
